// Win32Functions - native method interface DLL to selected Win32 functions

#include <windows.h>
#include <iostream>
#include <stdio.h>
#include "DirectXVideoOutputDevice.h"

using namespace std;

//function prototypes
DWORD WINAPI CreateDirectXWindow(LPVOID param);
LRESULT CALLBACK MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam);

#define WM_INITDIRECTX WM_APP+1
#define WM_EXIT        WM_APP+2

struct CUSTOMVERTEX
{
    FLOAT       x, y, z, rhw;
    FLOAT       tu, tv;
};

//the window class
WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC | CS_VREDRAW | CS_HREDRAW,
        MsgProc, 0L, 0L, GetModuleHandle(NULL), NULL,
        LoadCursor(NULL, IDC_ARROW) , NULL, NULL, "BlissDirectXWindow",
        NULL };

#define D3DFVF_CUSTOMVERTEX (D3DFVF_XYZRHW|D3DFVF_TEX1)

BYTE g_P8Palette[32] = {
    0x00, 0x01, 0x02, 0x03,
    0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0A, 0x0B,
    0x0C, 0x0D, 0x0E, 0x0F,
    0x00, 0x01, 0x02, 0x03,
    0x04, 0x05, 0x06, 0x07,
    0x08, 0x09, 0x0A, 0x0B,
    0x0C, 0x0D, 0x0E, 0x0F,
};

WORD g_R5G6B5Palette[32] = {
    0x0000, 0x017F, 0xF9E1, 0xC694,
    0x3368, 0x054B, 0xF749, 0xFFDF,
    0xBD58, 0x2DBF, 0xFD84, 0x5B80,
    0xFA8B, 0xA4DF, 0x7E4F, 0xB8EA,
    0x0000, 0x017F, 0xF9E1, 0xC694,
    0x3368, 0x054B, 0xF749, 0xFFDF,
    0xBD58, 0x2DBF, 0xFD84, 0x5B80,
    0xFA8B, 0xA4DF, 0x7E4F, 0xB8EA,

/* OLD PALETTE
    0x0000, 0x017F, 0xF9E2, 0xCE75,
    0x3B47, 0x052A, 0xFF4A, 0xFFFF,
    0xBD79, 0x25DF, 0xFDA3, 0x5360,
    0xFA6A, 0xA4BF, 0x7670, 0xB0EB,
    0x0000, 0x017F, 0xF9E2, 0xCE75,
    0x3B47, 0x052A, 0xFF4A, 0xFFFF,
    0xBD79, 0x25DF, 0xFDA3, 0x5360,
    0xFA6A, 0xA4BF, 0x7670, 0xB0EB,
*/
};

WORD g_X1R5G5B5Palette[32] = {
    0x0000, 0x00DF, 0x7D01, 0x6354,
    0x19A8, 0x02AB, 0x7B89, 0x7FFF,
    0x5EB8, 0x16DF, 0x7EC4, 0x2DC0,
    0x7D4B, 0x527F, 0x3F2F, 0x5C6A,
    0x0000, 0x00DF, 0x7D01, 0x6354,
    0x19A8, 0x02AB, 0x7B89, 0x7FFF,
    0x5EB8, 0x16DF, 0x7EC4, 0x2DC0,
    0x7D4B, 0x527F, 0x3F2F, 0x5C6A,

/* OLD PALETTE
    0x0000, 0x00DF, 0x7D02, 0x6755,
    0x1DA8, 0x02AB, 0x7FAA, 0x7FFF,
    0x62D9, 0x16FF, 0x7EE4, 0x2DC0,
    0x7D4B, 0x567F, 0x3F50, 0x5C6B,
    0x0000, 0x00DF, 0x7D02, 0x6755,
    0x1DA8, 0x02AB, 0x7FAA, 0x7FFF,
    0x62D9, 0x16FF, 0x7EE4, 0x2DC0,
    0x7D4B, 0x567F, 0x3F50, 0x5C6B,
*/
};

PALETTEENTRY g_A8R8G8B8Palette[32] = {
    { 0x00, 0x00, 0x00, 0xFF }, //BLACK
    { 0x00, 0x2D, 0xFF, 0xFF }, //BLUE
    { 0xFF, 0x3D, 0x10, 0xFF }, //RED
    { 0xC9, 0xCF, 0xAB, 0xFF }, //TAN
    { 0x38, 0x6B, 0x3F, 0xFF }, //DARK GREEN
    { 0x00, 0xA7, 0x56, 0xFF }, //GREEN
    { 0xFA, 0xEA, 0x50, 0xFF }, //YELLOW
    { 0xFF, 0xFC, 0xFF, 0xFF }, //WHITE
    { 0xBD, 0xAC, 0xC8, 0xFF }, //GRAY
    { 0x24, 0xB8, 0xFF, 0xFF }, //CYAN
    { 0xFF, 0xB4, 0x1F, 0xFF }, //ORANGE
    { 0x54, 0x6E, 0x00, 0xFF }, //BROWN
    { 0xFF, 0x4E, 0x57, 0xFF }, //PINK
    { 0xA4, 0x96, 0xFF, 0xFF }, //LIGHT BLUE
    { 0x75, 0xCC, 0x80, 0xFF }, //YELLOW GREEN
    { 0xB5, 0x1A, 0x58, 0xFF }, //PURPLE
};

DirectXVideoOutputDevice* currentDevice;

//potential error messages initializing DirectX
INT32 ERR_UNABLE_TO_INIT_WINDOW                = 1;
INT32 ERR_UNABLE_TO_RETRIEVE_VIDEO_MODE        = 2;
INT32 ERR_UNABLE_TO_SET_REQUESTED_VIDEO_MODE   = 3;
INT32 ERR_REQUEST_FOR_IN_MEMORY_TEXTURE_FAILED = 4;
INT32 ERR_REQUEST_FOR_HARDWARE_TEXTURE_FAILED  = 5;
INT32 ERR_REQUEST_FOR_VERTEX_BUFFER_FAILED     = 6;
INT32 ERR_UNABLE_TO_LOCK_VERTEX_BUFFER         = 7;
INT32 ERR_REQUEST_TO_CLEAR_BACK_BUFFER_FAILED  = 8;

const INT32 DirectXVideoOutputDevice::DEFAULT_WINDOW_WIDTH = 640;
const INT32 DirectXVideoOutputDevice::DEFAULT_WINDOW_HEIGHT = 384;

DirectXVideoOutputDevice::DirectXVideoOutputDevice()
    : modes(NULL),
      modeCount(0),
      fullScreen(FALSE),
      defaultFullScreenMode(NULL)
{
    getDisplayModes();
}

void DirectXVideoOutputDevice::setFullScreenMode(BOOL b)
{
    if (b && defaultFullScreenMode == NULL)
        return;

    fullScreen = b;
}

INT32 DirectXVideoOutputDevice::init(Intellivision* inty)
{
    //this becomes the current device
    currentDevice = this;

    //register the window class
    RegisterClassEx(&wc);

    //create the synchronization event
    m_hEventObject = CreateEvent(NULL, TRUE, FALSE, NULL);

    //capture a reference to the intellivision
    this->inty = inty;

    //capture the filter setting
    //filtered =  filtered;

    //reset the event we'll use to sync with the DirectX thread
    ResetEvent(m_hEventObject);

    //spin off a thread to create the new window and initialize DirectX
    m_hDirectXThreadHandle = CreateThread(NULL, NULL,
            CreateDirectXWindow, NULL, NULL, &m_directXThreadID);
    
    //wait for DirectX to complete initialization or the thread to exit
    HANDLE objects[2] = { m_hDirectXThreadHandle, m_hEventObject };
    WaitForMultipleObjects(2, objects, FALSE, INFINITE);

    //make sure the thread got off to a good start
    DWORD status;
    GetExitCodeThread(m_hDirectXThreadHandle, &status);
    if (status != STILL_ACTIVE)
        return status;

    //set the window title
    SendMessage(m_hWnd, WM_SETTEXT, 0, (LPARAM)inty->getCartridge()->getName());

    //everything's cool, fool.
    return 0;
}

void DirectXVideoOutputDevice::displayImage(UINT8* image)
{
    //capture a global hook onto the image data
    this->imageData = image;

    SendMessage(m_hWnd, WM_PAINT, 0, 0);
}

INT32 DirectXVideoOutputDevice::copyAndDisplayImage() {
    if (m_pd3dDevice == NULL) {
        return 0;
    }

    //check if the device was lost and try to reset it if possible
    HRESULT result;
    if ((result = m_pd3dDevice->TestCooperativeLevel()) != D3D_OK) {
        if (result != D3DERR_DEVICENOTRESET) {
            return -1;
        }

        destroyResources();
        if (m_pd3dDevice->Reset(&presentation) != D3D_OK) {
            return -1;
        }

        if (createResources() != 0) {
            return -1;
        }

        if (m_pd3dDevice->TestCooperativeLevel() != D3D_OK) {
            return -1;
        }
    }

    //if there is a new image available, copy it to our offscreen texture
    if (imageData != NULL) {
        //lock the rectangle in the texture that we're going to update
        D3DLOCKED_RECT locked;
        RECT rect = { 0, 0, 160, 192 };
        if (FAILED(m_pMemTexture->LockRect(0, &locked, &rect, 0))) {
            return -1;
        }

        WORD* backBuffer = (WORD*)locked.pBits;
        //a design pattern call "loop unrolling" is intentionally used here
        //to increase performance a bit; please don't roll this loop back up
        INT32 x = 0;
        INT32 y = 0;
        for (INT32 i = 0; i < 61440; i += 64) {
            backBuffer[y+x]    = m_pMemPalette[imageData[i]];
            backBuffer[y+x+1]  = m_pMemPalette[imageData[i+2]];
            backBuffer[y+x+2]  = m_pMemPalette[imageData[i+4]];
            backBuffer[y+x+3]  = m_pMemPalette[imageData[i+6]];
            backBuffer[y+x+4]  = m_pMemPalette[imageData[i+8]];
            backBuffer[y+x+5]  = m_pMemPalette[imageData[i+10]];
            backBuffer[y+x+6]  = m_pMemPalette[imageData[i+12]];
            backBuffer[y+x+7]  = m_pMemPalette[imageData[i+14]];
            backBuffer[y+x+8]  = m_pMemPalette[imageData[i+16]];
            backBuffer[y+x+9]  = m_pMemPalette[imageData[i+18]];
            backBuffer[y+x+10] = m_pMemPalette[imageData[i+20]];
            backBuffer[y+x+11] = m_pMemPalette[imageData[i+22]];
            backBuffer[y+x+12] = m_pMemPalette[imageData[i+24]];
            backBuffer[y+x+13] = m_pMemPalette[imageData[i+26]];
            backBuffer[y+x+14] = m_pMemPalette[imageData[i+28]];
            backBuffer[y+x+15] = m_pMemPalette[imageData[i+30]];
            backBuffer[y+x+16] = m_pMemPalette[imageData[i+32]];
            backBuffer[y+x+17] = m_pMemPalette[imageData[i+34]];
            backBuffer[y+x+18] = m_pMemPalette[imageData[i+36]];
            backBuffer[y+x+19] = m_pMemPalette[imageData[i+38]];
            backBuffer[y+x+20] = m_pMemPalette[imageData[i+40]];
            backBuffer[y+x+21] = m_pMemPalette[imageData[i+42]];
            backBuffer[y+x+22] = m_pMemPalette[imageData[i+44]];
            backBuffer[y+x+23] = m_pMemPalette[imageData[i+46]];
            backBuffer[y+x+24] = m_pMemPalette[imageData[i+48]];
            backBuffer[y+x+25] = m_pMemPalette[imageData[i+50]];
            backBuffer[y+x+26] = m_pMemPalette[imageData[i+52]];
            backBuffer[y+x+27] = m_pMemPalette[imageData[i+54]];
            backBuffer[y+x+28] = m_pMemPalette[imageData[i+56]];
            backBuffer[y+x+29] = m_pMemPalette[imageData[i+58]];
            backBuffer[y+x+30] = m_pMemPalette[imageData[i+60]];
            backBuffer[y+x+31] = m_pMemPalette[imageData[i+62]];
            x += 32;
            if (x == 160) {
                x = 0;
                y += m_pMemDesc.Width;
            }
        }

        if (FAILED(m_pMemTexture->UnlockRect(0))) {
            return -1;
        }

        //since the copying is done, release our reference to the image
        imageData = NULL;
    }

    //now render the image onto the device
    if (FAILED(m_pd3dDevice->BeginScene())) {
        return -1;
    }

    //update the hardware texture
    if (FAILED(m_pd3dDevice->UpdateTexture(
            m_pMemTexture, m_pCardTexture)))
    {
        return -1;
    }

    if (FAILED(m_pd3dDevice->SetTexture(0, m_pCardTexture)))
    {
        return -1;
    }
    if (FAILED(m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLOROP,
            D3DTOP_SELECTARG1)))
    {
        return -1;
    }
    if (FAILED(m_pd3dDevice->SetTextureStageState(0, D3DTSS_COLORARG1,
            D3DTA_TEXTURE)))
    {
        return -1;
    }
    if (filtered) {
        if (FAILED(m_pd3dDevice->SetTextureStageState(0, D3DTSS_MAGFILTER,
                D3DTEXF_LINEAR)))
        {
            return -1;
        }
    }
    if (FAILED(m_pd3dDevice->SetStreamSource(0, m_pVB, sizeof(CUSTOMVERTEX))))
    {
        return -1;
    }
    if (FAILED(m_pd3dDevice->SetVertexShader(D3DFVF_CUSTOMVERTEX)))
    {
        return -1;
    }
    if (FAILED(m_pd3dDevice->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2)))
    {
        return -1;
    }
    if (FAILED(m_pd3dDevice->SetTexture(0, NULL))) {
        return -1;
    }
    if (FAILED(m_pd3dDevice->EndScene())) {
        return -1;
    }

    if (FAILED(m_pd3dDevice->Present(NULL, NULL, NULL, NULL))) {
        return -1;
    }

    return 0;
} 

void DirectXVideoOutputDevice::release(Intellivision* inty) {
    //send message to destroy the DirectX window
    PostMessage(m_hWnd, WM_EXIT, 0, 0);

    //wait for the thread to exit
    WaitForSingleObject(m_hDirectXThreadHandle, INFINITE);

    //For some very strange reason, when returning from full screen mode on
    //Win2k, the system locks up unless we sleep a little, presumably because
    //the Win16Mutex has not yet been released, but why???  DirectX is done
    //shutting down so the resolution change should be completed by now.
    //If anyone reads this and can explain what the heck is the reason for
    //this, please let me know.  KLD
    if (fullScreen)
        Sleep(500);

    UnregisterClass(wc.lpszClassName, wc.hInstance);
 
    //delete the thread handle
    CloseHandle(m_hDirectXThreadHandle);
} 

DWORD WINAPI CreateDirectXWindow(LPVOID param)
{
    //initialize the window
    INT32 status = currentDevice->initWindow();
    if (status != 0)
        return status;

    PostMessage(currentDevice->m_hWnd, WM_INITDIRECTX, 0, 0);

    //start the message pump
    MSG	msg;
    while(GetMessage(&msg, NULL, 0, 0)) {
        TranslateMessage(&msg);
        DispatchMessage(&msg);
    }

    //clean up DirectX
    currentDevice->destroyResources();
    currentDevice->destroyDirectX();

    //exit the thread
    return msg.wParam;
}

INT32 DirectXVideoOutputDevice::initWindow() {
    DWORD windowStyle = WS_OVERLAPPED |
            (!fullScreen ? WS_CAPTION | WS_THICKFRAME | WS_SYSMENU : 0);

    //center the window on the desktop
    RECT windowRect;
    HWND desktopWnd;
    desktopWnd = GetDesktopWindow();
    RECT desktopDimensions;
    GetWindowRect(desktopWnd, &desktopDimensions);
    windowRect.left = ((desktopDimensions.right-desktopDimensions.left)-
            (fullScreen ? defaultFullScreenMode->Width : DEFAULT_WINDOW_WIDTH)
            )/2;
    windowRect.top = ((desktopDimensions.bottom-desktopDimensions.top)-
            (fullScreen ? defaultFullScreenMode->Height: DEFAULT_WINDOW_HEIGHT)
            )/2;
    windowRect.right = windowRect.left +
            (fullScreen ? defaultFullScreenMode->Width : DEFAULT_WINDOW_WIDTH);
    windowRect.bottom = windowRect.top +
            (fullScreen ? defaultFullScreenMode->Height: DEFAULT_WINDOW_HEIGHT);
    if (!fullScreen)
        AdjustWindowRect(&windowRect, windowStyle, FALSE);

    //create a window of the class
    m_hWnd = CreateWindow("BlissDirectXWindow",
            TEXT(""), windowStyle,
            windowRect.left, windowRect.top,
            windowRect.right-windowRect.left,
            windowRect.bottom-windowRect.top,
            NULL, NULL, GetModuleHandle(NULL), NULL);

    if (m_hWnd == NULL)
        return ERR_UNABLE_TO_INIT_WINDOW;

    //finally, show the window
    ShowWindow(m_hWnd, SW_SHOWNORMAL);
    BringWindowToTop(m_hWnd);
    UpdateWindow(m_hWnd);

    return 0;
}

INT32 DirectXVideoOutputDevice::initDirectX() {
    //initialize DirectX
    if((m_pD3D = Direct3DCreate8(D3D_SDK_VERSION)) == NULL) {
        return ERR_UNABLE_TO_INIT_WINDOW;
    }

    //get the current display mode
    D3DDISPLAYMODE d3ddm;
    if(FAILED(m_pD3D->GetAdapterDisplayMode(D3DADAPTER_DEFAULT,
            &d3ddm)))
    {
        m_pD3D->Release();
        m_pD3D = NULL;
        return ERR_UNABLE_TO_RETRIEVE_VIDEO_MODE;
    }

    //set up our request for a DirectX device
    ZeroMemory(&presentation, sizeof(presentation));
    presentation.BackBufferWidth =
            (fullScreen ? defaultFullScreenMode->Width : DEFAULT_WINDOW_WIDTH);
    presentation.BackBufferHeight =
            (fullScreen ? defaultFullScreenMode->Height: DEFAULT_WINDOW_HEIGHT);
    presentation.BackBufferFormat =
            (fullScreen ? defaultFullScreenMode->Format : d3ddm.Format);
    presentation.BackBufferCount = 1;
    presentation.MultiSampleType = D3DMULTISAMPLE_NONE;
    presentation.SwapEffect = D3DSWAPEFFECT_DISCARD;
    presentation.hDeviceWindow = m_hWnd;
    presentation.Windowed = !fullScreen;
    presentation.EnableAutoDepthStencil = FALSE;
    presentation.Flags = 0;
    presentation.FullScreen_RefreshRateInHz = D3DPRESENT_RATE_DEFAULT;
    presentation.FullScreen_PresentationInterval = D3DPRESENT_INTERVAL_DEFAULT;

    //request the DirectX device
    if(FAILED(m_pD3D->CreateDevice(D3DADAPTER_DEFAULT,
            D3DDEVTYPE_HAL, m_hWnd,
            D3DCREATE_SOFTWARE_VERTEXPROCESSING, &presentation,
            &m_pd3dDevice)))
    {
        m_pD3D->Release();
        m_pD3D = NULL;
        return ERR_UNABLE_TO_SET_REQUESTED_VIDEO_MODE;
    }

    return 0;
}

INT32 DirectXVideoOutputDevice::createResources() {
    //create our in-memory texture
    if (FAILED(D3DXCreateTexture(m_pd3dDevice, 160, 192, 1, 0,
            D3DFMT_X1R5G5B5, D3DPOOL_SYSTEMMEM, &m_pMemTexture)))
    {
        return ERR_REQUEST_FOR_IN_MEMORY_TEXTURE_FAILED;
    }

    if (FAILED(m_pMemTexture->GetLevelDesc(0, &m_pMemDesc))) {
        m_pMemTexture->Release();
        m_pMemTexture = NULL;
        return ERR_REQUEST_FOR_IN_MEMORY_TEXTURE_FAILED;
    }

    if (m_pMemDesc.Format == D3DFMT_X1R5G5B5)
        m_pMemPalette = g_X1R5G5B5Palette;
    else
        m_pMemPalette = g_R5G6B5Palette;

    //create the displayable texture on the video card
    if (FAILED(D3DXCreateTexture(m_pd3dDevice, 160, 192, 1, 0,
            D3DFMT_X1R5G5B5, D3DPOOL_DEFAULT, &m_pCardTexture)))
    {
        m_pMemTexture->Release();
        m_pMemTexture = NULL;
        return ERR_REQUEST_FOR_HARDWARE_TEXTURE_FAILED;
    }

    //create our vertex buffer
    float right = (float)
            (fullScreen ? defaultFullScreenMode->Width : DEFAULT_WINDOW_WIDTH);
    float bottom = (float)
            (fullScreen ? defaultFullScreenMode->Height: DEFAULT_WINDOW_HEIGHT);
    float rightTexPin = (160.2f/(float)m_pMemDesc.Width);
    float bottomTexPin = (192.3f/(float)m_pMemDesc.Height);
    CUSTOMVERTEX Vertices[4] =
    {
        {  0.0f,   0.0f, 0.0f, 0.1f, 0.0f, 0.0f },
        { right,   0.0f, 0.0f, 0.1f, rightTexPin,         0.0f },
        {  0.0f, bottom, 0.0f, 0.1f,        0.0f, bottomTexPin },
        { right, bottom, 0.0f, 0.1f, rightTexPin, bottomTexPin },
    };

    if (FAILED(m_pd3dDevice->CreateVertexBuffer(
            4*sizeof(CUSTOMVERTEX), 0, D3DFVF_CUSTOMVERTEX, D3DPOOL_DEFAULT,
            &m_pVB)))
    {
        m_pCardTexture->Release();
        m_pCardTexture = NULL;
        m_pMemTexture->Release();
        m_pMemTexture = NULL;
        return ERR_REQUEST_FOR_VERTEX_BUFFER_FAILED;
    }

    VOID* pVertices;
    if (FAILED(m_pVB->Lock(0, sizeof(Vertices),
            (BYTE**)&pVertices, 0)))
    {
        m_pVB->Release();
        m_pVB = NULL;
        m_pCardTexture->Release();
        m_pCardTexture = NULL;
        m_pMemTexture->Release();
        m_pMemTexture = NULL;
        return ERR_UNABLE_TO_LOCK_VERTEX_BUFFER;
    }
    memcpy(pVertices, Vertices, sizeof(Vertices));
    m_pVB->Unlock();

    return 0;
}

void DirectXVideoOutputDevice::destroyResources() {
    if (m_pVB) {
        m_pVB->Release();
        m_pVB = NULL;
    }
    if (m_pCardTexture) {
        m_pCardTexture->Release();
        m_pCardTexture = NULL;
    }
    if (m_pMemTexture) {
        m_pMemTexture->Release();
        m_pMemTexture = NULL;
    }
}

void DirectXVideoOutputDevice::destroyDirectX() {
    if (m_pd3dDevice) {
        m_pd3dDevice->Release();
        m_pd3dDevice = NULL;
    }
    if (m_pD3D) {
        m_pD3D->Release();
        m_pD3D= NULL;
    }
}

void DirectXVideoOutputDevice::getDisplayModes()
{
    //clear the current list of video modes
    if (modes != NULL) {
        delete[] modes;
        modes = NULL;
        modeCount = 0;
        defaultFullScreenMode = NULL;
    }
    
    //initialize DirectX
    if((m_pD3D = Direct3DCreate8(D3D_SDK_VERSION)) == NULL) {
        return;
    }

    UINT modeCount;
    INT32 totalModeCount = m_pD3D->GetAdapterModeCount(D3DADAPTER_DEFAULT);

    //get the list of all valid modes
    modes = new D3DDISPLAYMODE[totalModeCount];
    modeCount = 0;
    BOOL alreadyHaveIt;
    for (UINT i = 0; i < totalModeCount; i++) {
        m_pD3D->EnumAdapterModes(D3DADAPTER_DEFAULT, i, &modes[modeCount]);

        //ignore 32-bit modes.  even 16-bit is overkill for an Intellivision
        if (modes[modeCount].Format == D3DFMT_X8R8G8B8 ||
                modes[modeCount].Format == D3DFMT_A8R8G8B8)
            continue;

        //if this mode is 800x600, then we want to use it as the default
        //display mode
        if (modes[modeCount].Width == 800 && modes[modeCount].Height == 600)
            defaultFullScreenMode = &modes[modeCount];

        //check to see if we've already found a similar mode
        alreadyHaveIt = FALSE;
        for (UINT j = modeCount; j > 0; j--) {
            if (modes[j-1].Width == modes[modeCount].Width &&
                    modes[j-1].Height == modes[modeCount].Height)
            {
                alreadyHaveIt = TRUE;
                break;
            }
        }

        if (!alreadyHaveIt)
            modeCount++;
    }

    m_pD3D->Release();
    m_pD3D = NULL;
}

LRESULT CALLBACK MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    int status;
    switch (msg) {
        case WM_INITDIRECTX:
            //initialize DirectX
            status = currentDevice->initDirectX();
            if (status != 0) {
                DestroyWindow(currentDevice->m_hWnd);
                PostQuitMessage(status);
                return 0;
            }
        
            status = currentDevice->createResources();
            if (status != 0) {
                currentDevice->destroyDirectX();
                PostQuitMessage(status);
                return 0;
            }
        
            //clear the back buffer to black
            currentDevice->m_pd3dDevice->Clear(0, NULL, D3DCLEAR_TARGET,
                    D3DCOLOR_XRGB(0,0,0), 1.0f, 0L);
        
            //display the now cleared back buffer
            if (FAILED(currentDevice->m_pd3dDevice->Present(NULL, NULL, NULL,
                    NULL)))
            {
                PostQuitMessage(ERR_REQUEST_TO_CLEAR_BACK_BUFFER_FAILED);
                return 0;
            }
            ValidateRect(currentDevice->m_hWnd, NULL);
        
            //signal that initialization was successfully completed
            SetEvent(currentDevice->m_hEventObject);

            return 0;
        
        case WM_CLOSE:
            if (currentDevice->inty != NULL)
                currentDevice->inty->turnOff();
            return 0;

        case WM_PAINT:
            currentDevice->copyAndDisplayImage();
            ValidateRect(hWnd, NULL);
            return 0;

        case WM_DESTROY:
            PostQuitMessage(0);
            return 0;

        case WM_EXIT:
            DestroyWindow(hWnd);
            return 0;
    }
    return DefWindowProc(hWnd, msg, wParam, lParam);
}

